import os
import sys

import traceback

from dew.exceptions import DewfileError


class View(object):

    def __init__(self):
        self.show_verbose = False

    def info(self, message):
        print(message)

    def verbose(self, message):
        if self.show_verbose:
            print(message)

    def error(self, message):
        sys.stderr.write(message)
        sys.stderr.write(os.linesep)

    def dewfile_error(self, e: DewfileError) -> None:
        message = ''.join(traceback.format_exception(type(e.__cause__), e.__cause__, e.inner_tb))
        self.error(message)
        self.error(f'Failed to parse dewfile at path {e.file_path}')





class DewError(Exception):
    pass


class PullError(DewError):
    pass


class BuildError(DewError):
    pass


class DewfileError(DewError):
    def __init__(self, file_path: str, inner_tb) -> None:
        self.file_path = file_path
        self.inner_tb = inner_tb
        super().__init__()



import os.path
import shutil
from typing import Callable


class StorageController(object):

    def __init__(self, path):
        self.path = path

    def ensure_directories_exist(self):
        self.walk_directories(lambda p: os.makedirs(p, exist_ok=True))

    def clean(self):
        self.walk_directories(lambda p: shutil.rmtree(p))

    def walk_directories(self, f: Callable[[str], None]):
        f(self.get_storage_dir())
        f(self.get_sources_dir())
        f(self.get_builds_dir())
        f(self.get_downloads_dir())
        f(self.get_install_dir())

    def get_storage_dir(self) -> str:
        return self.join_storage_dir_path()

    def get_sources_dir(self) -> str:
        return self.join_storage_dir_path('sources')

    def get_builds_dir(self) -> str:
        return self.join_storage_dir_path('builds')

    def get_build_dir(self, label: str) -> str:
        return os.path.join(self.get_builds_dir(), label)

    def get_downloads_dir(self) -> str:
        return self.join_storage_dir_path('downloads')

    def get_install_dir(self) -> str:
        return self.join_storage_dir_path('prefix')

    def get_output_prefix_dir(self) -> str:
        return self.join_storage_dir_path('output-prefixes')

    def join_storage_dir_path(self, *args) -> str:
        return os.path.join(self.path, *args)



import argparse
import dew.command
from dew.dependencyprocessor import DependencyProcessor
from dew.dewfile import Dependency
from dew.impl import CommandData


class ArgumentData(object):
    def __init__(self) -> None:
        self.name = ''


class Command(dew.command.Command):

    def get_argparser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('name')
        return parser

    def execute(self, args: ArgumentData, data: CommandData) -> int:
        dewfile = data.project_parser.parse()

        target_dep: Dependency = None
        for dep in dewfile.dependencies:
            if dep.name == args.name:
                target_dep = dep

        if target_dep is None:
            data.view.error(f'Unknown dependency "{args.name}"')
            return 1

        previous_ref = target_dep.ref

        processor = DependencyProcessor(data.storage, data.view, target_dep, dewfile, data.properties)
        target_dep.ref = processor.get_remote().get_latest_ref()
        data.project_parser.save(dewfile)

        if previous_ref != target_dep.ref:
            data.view.info(
                f'Dependency {args.name} upgraded.\n'
                f'Head:         {target_dep.head}\n'
                f'Previous ref: {previous_ref}\n'
                f'New ref:      {target_dep.ref}'
            )
        else:
            data.view.info(f'Dependency {args.name} is already at latest ref {previous_ref} for head {target_dep.head}')

        return 0



import argparse

from dew.args import ArgumentData
from dew.impl import CommandData
from dew.projectproperties import ProjectProperties


class Command(object):
    def get_argparser(self) -> argparse.ArgumentParser:
        pass

    def set_properties_from_args(self, args: ArgumentData, properties: ProjectProperties) -> None:
        pass

    def execute(self, args: ArgumentData, data: CommandData) -> int:
        pass

    def cleanup(self, args: ArgumentData, data: CommandData) -> None:
        pass



import os

from dew.builder import Builder
from dew.projectproperties import ProjectProperties
from dew.dewfile import Dependency
from dew.exceptions import BuildError
from dew.subprocesscaller import SubprocessCaller
from dew.view import View


class XcodeBuilder(Builder):
    def __init__(self, buildfile_dir: str, build_dir: str, install_dir: str, dependency: Dependency,
                 options: ProjectProperties, caller: SubprocessCaller, view: View) -> None:
        self.buildfile_dir = buildfile_dir
        self.build_dir = build_dir
        self.install_dir = install_dir
        self.dependency = dependency
        self.options = options
        self.caller = caller
        self.view = view

    def build(self) -> None:
        build_dir = self.build_dir
        os.makedirs(build_dir, exist_ok=True)

        # Figure out which xcodeproj path we are using
        xcodeproj_path = ''
        for path in os.listdir(os.path.join(self.buildfile_dir)):
            if path.endswith('.xcodeproj'):
                xcodeproj_path = os.path.join(self.buildfile_dir, path)
                break

        if not xcodeproj_path:
            self.view.error('Cannot determine the xcodeproj to build with.')
            raise BuildError()

        self.caller.call(
            [
                'xcodebuild', '-project', xcodeproj_path,
                'OBJROOT=' + os.path.join(build_dir, 'Intermediates'),
                'BUILD_DIR=' + os.path.join(build_dir, 'Products'),
                'SYMROOT=' + os.path.join(build_dir, 'Products'),
                'build'
            ],
            cwd=self.build_dir
        )



import os
import argparse
import shutil

import dew.command
from dew.impl import CommandData


class ArgumentData(object):
    def __init__(self) -> None:
        pass


class Command(dew.command.Command):
    def get_argparser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(add_help=False)
        return parser

    def execute(self, args: ArgumentData, data: CommandData) -> int:
        src_path = os.path.join(__file__, '..', '..', '..', 'cmake', 'dew.cmake')
        src_path = os.path.normpath(src_path)
        dest_path = os.path.join(os.getcwd(), 'dew.cmake')
        shutil.copy(src_path, dest_path)
        return 0



import argparse

import dew.command
from dew.dependencyprocessor import DependencyProcessor
from dew.dewfile import Dependency
from dew.impl import CommandData


class ArgumentData(object):
    def __init__(self) -> None:
        self.name = ''


class Command(dew.command.Command):
    def get_argparser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('name', help='Name of the dependency to finish local work for')
        return parser

    def execute(self, args: ArgumentData, data: CommandData) -> int:
        dewfile = data.project_parser.parse()

        target_dep: Dependency = None
        for dep in dewfile.dependencies:
            if dep.name == args.name:
                target_dep = dep

        if target_dep is None:
            data.view.error(f'Unknown dependency "{args.name}"')
            return 1

        if args.name not in dewfile.local_overrides:
            data.view.error(f'Local work on {args.name} has not been started.')
            return 1

        local_path = dewfile.local_overrides[args.name]

        processor = DependencyProcessor(data.storage, data.view, target_dep, dewfile, data.properties, source_dir=local_path)

        if processor.get_remote().has_pending_changes():
            data.view.error(f'Local {args.name} has pending changes. Please resolve.')
            return 1

        new_ref = processor.get_remote().get_current_ref()
        new_head = processor.get_remote().get_current_head()
        old_ref = target_dep.ref
        old_head = target_dep.head

        target_dep.ref = new_ref
        target_dep.head = new_head

        del dewfile.local_overrides[args.name]
        data.project_parser.save(dewfile)
        data.project_parser.save_local_work(dewfile)

        if new_ref != old_ref:
            data.view.info(
                f'Previous ref: {old_ref}\n'
                f'New ref:      {new_ref}'
            )
        if new_head != old_head:
            data.view.info(
                f'Previous head: {old_head}\n'
                f'New head:      {new_head}'
            )

        return 0



import sys
import os
import json
from typing import List, Any, Dict

from dew.exceptions import DewfileError, DewError


class DewFile(object):
    def __init__(self) -> None:
        self.dependencies: List[Dependency] = []
        self.local_overrides: Dict[str, str] = {}


class Dependency(object):
    def __init__(self) -> None:
        self.name = ''
        self.url = ''
        self.type = ''
        self.head = ''
        self.ref = ''
        self.buildfile_dir = ''


def _parse_dewfile(data: Dict[str, Any]) -> DewFile:
    dependencies = []
    dependencies_obj = data['dependencies']

    for dep in dependencies_obj:
        dependencies.append(parse_dependency(dep))

    dewfile = DewFile()
    dewfile.dependencies = dependencies
    return dewfile


def serialize_dewfile(dewfile: DewFile) -> Dict[str, Any]:
    data = {}
    dependencies = []
    for dep in dewfile.dependencies:
        dependencies.append(serialize_dependency(dep))
    data['dependencies'] = dependencies
    return data


def parse_dependency(obj: Dict[str, Any]) -> Dependency:
    dep = Dependency()
    dep.name = obj['name']
    dep.url = obj['url']
    dep.type = obj['type']
    dep.head = obj['head']
    dep.ref = obj.get('ref', '')
    dep.buildfile_dir = obj.get('buildfile_dir', '')
    return dep


def serialize_dependency(dep: Dependency) -> Dict[str, Any]:
    data = {
        'name': dep.name,
        'url': dep.url,
        'type': dep.type,
        'head': dep.head
    }

    if dep.ref:
        data['ref'] = dep.ref
    if dep.buildfile_dir:
        data['buildfile_dir'] = dep.buildfile_dir

    return data


def parse_dewfile(path: str) -> DewFile:
    try:
        dewfile_data = load_json(path)
        dewfile = _parse_dewfile(dewfile_data)
    except Exception as e:
        raise DewfileError(path, sys.exc_info()[2]) from e
    return dewfile


def parse_local_work_file(path: str) -> Dict[str, str]:
    if not os.path.isfile(path):
        return {}

    with open(path) as f:
        try:
            data = json.load(f)
        except Exception as e:
            raise DewfileError(path, sys.exc_info()[2]) from e

    if not isinstance(data, dict):
        raise DewError('fContents of {path} is misformatted.')

    return {name: str(path) for name, path in data.items()}


def save_dewfile(dewfile: DewFile, path: str) -> None:
    with open(path, 'w') as f:
        json.dump(serialize_dewfile(dewfile), f, indent=4)


def save_local_work(dewfile: DewFile, path: str) -> None:
    with open(path, 'w') as f:
        json.dump(dewfile.local_overrides, f, indent=4)


def load_json(path: str) -> Dict[str, Any]:
    with open(path) as file:
        return json.load(file)


class ProjectFilesParser(object):
    def __init__(self, dewfile_path: str):
        self.dewfile_path = dewfile_path

        path_without_ext, ext = os.path.splitext(self.dewfile_path)
        self.local_work_path = path_without_ext + '.local' + ext

    def parse(self) -> DewFile:
        dewfile = parse_dewfile(self.dewfile_path)
        dewfile.local_overrides = parse_local_work_file(self.local_work_path)
        return dewfile

    def save(self, dewfile: DewFile) -> None:
        save_dewfile(dewfile, self.dewfile_path)

    def save_local_work(self, dewfile: DewFile) -> None:
        save_local_work(dewfile, self.local_work_path)



import os

import argparse
import shutil

import dew.command
import dew.args
from dew.dependencyprocessor import DependencyProcessor
from dew.dewfile import Dependency
from dew.impl import CommandData


class ArgumentData(object):
    def __init__(self) -> None:
        self.name = ''
        self.path = ''
        self.existing = False


class Command(dew.command.Command):

    def get_argparser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('name', help='Name of the dependency to work on locally')
        parser.add_argument('path', help='Path to the directory where the dependency source will be put or is located')
        parser.add_argument('--existing', action='store_true',
                            help='Treat given path as an existing source directory of the given dependency')
        return parser

    def execute(self, args: ArgumentData, data: CommandData) -> int:
        dewfile = data.project_parser.parse()

        target_dep: Dependency = None
        for dep in dewfile.dependencies:
            if dep.name == args.name:
                target_dep = dep

        if target_dep is None:
            data.view.error(f'Unknown dependency "{args.name}"')
            return 1

        if args.name in dewfile.local_overrides:
            data.view.error(f'Local work on {args.name} has already been started.')
            return 1

        if not args.existing and os.path.exists(args.path) and len(os.listdir(args.path)) > 0:
            data.view.error('Target path is not empty. '
                            'Use the --existing flag to use this existing contents of this path.')
            return 1

        processor = DependencyProcessor(data.storage, data.view, target_dep, dewfile, data.properties)
        processor.get_remote().pull()

        if not args.existing:
            if os.path.exists(args.path):
                os.rmdir(args.path)
            shutil.copytree(processor.get_default_source_dir(), args.path)

        dewfile.local_overrides[args.name] = args.path
        data.project_parser.save_local_work(dewfile)
        return 0



import argparse

import dew.command
from dew.impl import CommandData


class ArgumentData(object):
    def __init__(self) -> None:
        pass


class Command(dew.command.Command):
    def get_argparser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(add_help=False)
        return parser

    def execute(self, args: ArgumentData, data: CommandData) -> int:
        data.view.error('Not implemented yet')
        return 1



import multiprocessing
import os
from typing import Iterable, Optional

from dew.exceptions import BuildError
from dew.subprocesscaller import SubprocessCaller
from dew.builder import Builder
from dew.projectproperties import ProjectProperties
from dew.dewfile import Dependency
from dew.view import View


class CMakeBuilder(Builder):
    def __init__(self, buildfile_dir: str, build_dir: str, install_dir: str, dependency: Dependency,
                 properties: ProjectProperties, caller: SubprocessCaller, view: View, prefix_paths: Iterable[str]) -> None:
        self.buildfile_dir = os.path.abspath(buildfile_dir)
        self.build_dir = os.path.abspath(build_dir)
        self.install_dir = os.path.abspath(install_dir)
        self.dependency = dependency
        self.properties = properties
        self.caller = caller
        self.view = view
        self.prefix_paths = list(prefix_paths)

    def get_cmake_executable(self) -> str:
        cmake_executable = self.properties.cmake_executable
        if not cmake_executable:
            cmake_executable = 'cmake'
        return cmake_executable

    def build(self) -> None:
        install_dir = self.install_dir
        os.makedirs(self.build_dir, exist_ok=True)

        cmake_executable = self.get_cmake_executable()

        module_paths = [os.path.join(prefix, 'share', 'cmake', 'Modules') for prefix in self.prefix_paths]

        generator = self.properties.cmake_generator
        if not generator:
            generator = guess_generator()
            if not generator:
                raise BuildError('Cannot guess CMake generator to use. ' 
                                 'Please specify a generator using the --guess-generator flag.')

        args = [
            cmake_executable,
            '-G', generator,
            self.buildfile_dir,
            '-DCMAKE_INSTALL_PREFIX={0}'.format(install_dir),
            '-DCMAKE_PREFIX_PATH={0}'.format(';'.join(self.prefix_paths)),
            '-DCMAKE_MODULE_PATH={0}'.format(';'.join(module_paths)),
            '-DCMAKE_BUILD_TYPE=Debug',  # TODO: Support building dependencies in release mode.

            # Lots of projects default to build shared libs instead of static. Let's add some consistency.
            # TODO: Need a system to choose the types of libraries desired.
            '-DBUILD_SHARED_LIBS=OFF'
        ]

        # Setup environment
        env = {
            'PATH': os.environ.get('PATH'),
            'CC': self.properties.c_compiler_path,
            'CXX': self.properties.cxx_compiler_path
        }

        # Configure
        self.caller.call(args, cwd=self.build_dir, error_exception=BuildError, env=env)

        build_args = [cmake_executable, '--build', '.']
        if generator in ('Unix Makefiles', 'MinGW Makefiles'):
            build_args.extend(('--', '-j', str(multiprocessing.cpu_count())))

        # Build
        self.caller.call(
            build_args,
            cwd=self.build_dir,
            error_exception=BuildError
        )

        # Install
        self.caller.call(
            [cmake_executable, '--build', '.', '--target', 'install'],
            cwd=self.build_dir,
            error_exception=BuildError
        )


GUESSED_GENERATOR: Optional[str] = None


def guess_generator() -> Optional[str]:
    global GUESSED_GENERATOR
    if GUESSED_GENERATOR:
        return GUESSED_GENERATOR

    guess = None

    paths = os.environ.get('PATH', '').split(os.pathsep)
    if os.name == 'nt':
        extensions = os.environ.get('PATHEXT').split(';')
    else:
        extensions = ['']

    def locate(name: str) -> bool:
        for path in paths:
            for entry in os.listdir(path):
                filename, ext = os.path.splitext(entry)
                if filename != name:
                    continue
                if ext not in extensions:
                    continue
                return True

    if locate('make'):
        guess = 'Unix Makefiles'
    elif locate('mingw32-make'):
        guess = 'MinGW Makefiles'

    GUESSED_GENERATOR = guess
    return guess



import os

from dew.dewfile import Dependency
from dew.remote import Remote


class LocalRemote(Remote):
    def __init__(self, dependency: Dependency, dest_dir: str) -> None:
        self.dependency = dependency

    def pull(self) -> None:
        pass

    def get_latest_ref(self) -> str:
        return str(int(self._get_latest_modtime()))

    def _get_latest_modtime(self) -> int:
        # Return unix timestamp of most recent file modification time in source directory
        return max(os.path.getmtime(root) for root, _, _ in os.walk(self.dependency.url))

    def get_current_ref(self) -> str:
        return self.get_latest_ref()

    def get_current_head(self) -> str:
        return self.dependency.head

    def has_pending_changes(self) -> bool:
        try:
            current_modtime = int(self.dependency.ref)
        except ValueError:
            return True
        return self._get_latest_modtime() > current_modtime

    def get_source_dir(self) -> str:
        return self.dependency.url



import os
import argparse
from typing import List, Optional

import dew.command
from dew.depstate import DependencyStateController
from dew.impl import CommandData
from dew.projectprocessor import ProjectProcessor
from dew.projectproperties import ProjectProperties


class ArgumentData(object):
    def __init__(self) -> None:
        self.cmake_generator = ''
        self.cmake_executable = ''
        self.c_compiler_path = ''
        self.cxx_compiler_path = ''
        self.additional_prefix_paths: List[str] = []


class Command(dew.command.Command):
    def __init__(self) -> None:
        self.depstates: Optional[DependencyStateController] = None

    def get_argparser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('--CC', dest='c_compiler_path', help='Path to the C compiler')
        parser.add_argument('--CXX', dest='cxx_compiler_path', help='Path to the CXX compiler')
        parser.add_argument('--prefix', dest='additional_prefix_paths', action='append')
        parser.add_argument('--cmake-generator')
        parser.add_argument('--cmake-executable')
        return parser

    def set_properties_from_args(self, args: ArgumentData, properties: ProjectProperties) -> None:
        if args.cmake_generator:
            properties.cmake_generator = args.cmake_generator
        if args.cmake_executable:
            properties.cmake_executable = args.cmake_executable
        if args.c_compiler_path:
            properties.c_compiler_path = args.c_compiler_path
        if args.cxx_compiler_path:
            properties.cxx_compiler_path = args.cxx_compiler_path
        if args.additional_prefix_paths:
            properties.prefixes = args.additional_prefix_paths

    def execute(self, args: ArgumentData, data: CommandData) -> int:
        dewfile = data.project_parser.parse()
        depstates = DependencyStateController(data.storage)
        self.depstates = depstates
        depstates.load()

        is_dirty = data.properties.dirty
        if is_dirty:
            data.view.info('Dew project properties have changed. Clearing dependency state information.')
            depstates.clear()
            depstates.save()

        data.properties.save()
        properties = data.properties.get()

        project_processor = ProjectProcessor(data.storage, properties, data.view, depstates)
        project_processor.set_data(dewfile)

        dewfile, refs_have_changed = project_processor.update_refs()
        if refs_have_changed:
            data.project_parser.save_refs(dewfile)

        project_processor.process()

        dummy_file_path = data.args.update_dummy_file
        if dummy_file_path:
            with open(dummy_file_path, 'a'):
                os.utime(dummy_file_path, None)

        return 0

    def cleanup(self, args: ArgumentData, data: CommandData) -> None:
        if self.depstates:
            self.depstates.save()



from dew import git
from dew.dewfile import Dependency
from dew.remote import Remote


class GitRemote(Remote):

    def __init__(self, dependency: Dependency, dest_dir: str) -> None:
        self.dependency = dependency
        self.dest_dir = dest_dir

    def pull(self) -> None:
        repo, origin = git.get_repo(self.dependency.url, self.dest_dir)
        git.fetch(origin, self.dependency.head)
        git.checkout(repo, origin, self.dependency.head, self.dependency.ref)

    def get_latest_ref(self) -> str:
        repo, origin = git.get_repo(self.dependency.url, self.dest_dir)
        git.fetch(origin, self.dependency.head)
        return git.get_latest_ref(origin, self.dependency.head)

    def get_current_ref(self) -> str:
        repo, origin = git.get_repo(self.dependency.url, self.dest_dir)
        return repo.head.commit.hexsha

    def get_current_head(self) -> str:
        repo, origin = git.get_repo(self.dependency.url, self.dest_dir)
        return repo.active_branch.name

    def has_pending_changes(self) -> bool:
        repo, origin = git.get_repo(self.dependency.url, self.dest_dir)
        return repo.is_dirty(untracked_files=True)

    def get_source_dir(self) -> str:
        return self.dest_dir



import subprocess
from typing import List, Type, Optional, Dict

from dew.view import View


class SubprocessCallError(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class SubprocessCaller(object):
    def __init__(self, view: View) -> None:
        self.view = view

    def call(self, args: List[str], cwd: str, error_exception: Type[Exception],
             env: Optional[Dict[str, str]] = None) -> None:
        self.view.verbose('Calling subprocess: "{0}", cwd: {1}'.format(repr(args), repr(cwd)))
        proc = subprocess.run(
            args, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env,
            universal_newlines=True
        )

        self.view.verbose('Process output:\n{0}'.format(proc.stdout))
        if len(proc.stderr) > 0:
            self.view.error(proc.stderr)

        if proc.returncode is not 0:
            raise error_exception()
